# Memory Benchmark

The `MemoryBenchmark.ino` was compiled with each `FEATURE_*` and the flash
memory and static RAM sizes were recorded. The `FEATURE_BASELINE` selection is
the baseline, and its memory usage  numbers are subtracted from the subsequent
`FEATURE_*` memory usage.

**Version**: AceTime v1.11.2

**DO NOT EDIT**: This file was auto-generated using `make README.md`.

## How to Regenerate

To regenerate this README.md:

```
$ make clean_benchmarks
$ make benchmarks
$ make README.md
```

The `make benchmarks` target uses `collect.sh` script which calls `auniter.sh`
(https://github.com/bxparks/AUniter) to invoke the Arduino IDE programmatically.
It produces a `*.txt` file with the flash and ram usage information (e.g.
`nano.txt`). It now takes about 16 minutes to generate the `*.txt` files on my
quad-core Intel Core i7-3840QM CPU @ 2.80GHz laptop.

The `make README.md` command calls the `generated_readme.py` Python script which
generates this `README.md` file. The ASCII tables below are generated by the
`generate_table.awk` script, which takes each `*.txt` file and converts it to an
ASCII table.

## Library Size Changes

**v1.3:**
* The `BasicZoneManager` and `ExtendedZoneManager` classes were unified under a
  new parent interface `ZoneManager`. This seems to have caused the flash size
  to increase by around 1200 bytes on the AVR processors (Nano, Pro Micro),
  about 500 bytes on a SAMD, about 800 bytes on a ESP8266, 100 bytes on a ESP32,
  and 1400 bytes on a Teensy 3.2. The 8-bit processors suffer the most flash
  size increase proportional to their limited 32kB limit.
* Adding the `ZoneManager` interface simplifies a lot of the complexity with
  saving and restoring time zones using the `TimeZoneData` object, and I think
  it is worth the extra cost of flash size. The mitigating factor is that
  applications targetted towards 8-bit processors will normally have fixed
  number of timezones at compile time, so they can avoid using a `ZoneManager`,
  and avoid this penalty in flash size.

**v1.4.1+:**
* Removed the `ZoneInfo::transitionBufSize` field from the `ZoneInfo` struct,
  which saves 1 byte on 8-bit processors (none on 32-bit processors due to
  4-byte alignment). We save 266 bytes for `BasicZoneManager` and 386 bytes for
  `ExtendedZoneManager` when all the zones are loaded into the zone registry.
* Incorporated zoneName compression causes flash/ram usage to increase by
  ~250/120 bytes when using only 1-2 zones, but *decreases* flash consumption by
  1200-2400 bytes when all the zones are loaded into the `ZoneManager`.

**v1.5+:**
* Changing `ZoneProcessorCache::getType()` from a `virtual` to a non-virtual
  method saves 250-350 bytes of flash memory when using a `BasicZoneManager` or
  an `ExtendedZoneManager` on an 8-bit AVR processor. Unexpectedly, the flash
  memory consumption *increases* slightly (~0-50 bytes) for some ARM processors
  and the ESP32. Since those processors have far more flash memory, this seems
  like a good tradeoff.
* Changing `BasicZoneProcessor` and `ExtendedZoneProcessor` to be subclasses of
  the templatized `BasicZoneProcessorTemplate` and
  `ExtendedZoneProcessorTemplate` classes causes reduction of flash consumption
  by 250-400 bytes for 32-bit processors. Don't know why. (Very little
  difference for 8-bit AVR).
* Adding a `BrokerFactory` layer of indirection (to support more complex
  ZoneProcessors and Brokers) causes flash memory to go up by 100-200 bytes.

**v1.6:**
* Added support for `LinkRegistry` to `BasicZoneManager` and
  `ExtendedZoneManager`. This increases the flash memory usage by 150-500 bytes
  when using one of these classes due to the code required by `LinkRegistrar`. This extra cost is incurred even if the `LinkRegistry` is set to 0 elements. Each `LinkEntry` consumes 8 bytes (2 x `uint32_t`). So a
  `zonedb::kLinkRegistry` with 183 elements uses 1464 extra bytes of flash; a
  `zonedbx::kLinkRegistry` with 207 elements uses 1656 extra bytes.

**v1.7:**
* The virtual destructor on the `Clock` base class removed. This reduced the
  flash usage by 618 bytes on AVR processors , 328 bytes on the SAMD21, but only
  50-60 bytes on other 32-bit processors.
* The various `printShortNameTo()` or `printShortTo()` methods changed to
  replace the underscore in the zone names (e.g. `Los_Angeles`) with spaces
  (e.g. `Los Angeles`) to be more human friendly. This made little difference in
  the flash memory consumption, except on the ESP32 where it increased by
  200-300 bytes.

**v1.7.2**
* The `SystemClock::clockMillis()` is now non-virtual, using compile-time
  polymorphism through C++ template, and incorporating the same techniques from
  AceRoutine v1.3. Saves about 20-40 bytes of flash.

**v1.7.5:**
* `ExtendedZoneProcessor.compareTransitionToMatch()` was modified to
  detect an exact equality between a `Transition` and its `MatchingEra` if any
  of the 3 time stamp versions ('w', 's', 'u') are equal. Adds about 120-150
  bytes of flash on 8-bit and 32-bit processors. But removing
  `originalTransitionTime` from `Transition` decreases flash usage by about 20
  bytes.
* Upgrade ESP8266 Core from 2.7.4 to 3.0.2. Flash consumption increases by
  3-5 kB across the board.
* Upgrade Teensyduino from 1.54 to 1.55. Add memory consumed by `malloc()` and
  `free()` when using classes with virtual methods into baseline
  MemoryBenchmark, reducing the actual memory usage of various features by
  ~3kB.

**v1.8.0:**
* Move Clock and SystemClock benchmarks into AceTimeClock v1.0.0.
* Extract thin links from BasicZoneManager and ExtendedZoneManager into
  new BasicLinkManager and ExtendedLinkManager classes.
    * Saves 200-500 bytes of flash for BasicZoneManager and ExtendedZoneManager.
    * Applications can decide whether to use thin links through the LinkManager
      (~2000 flash bytes for AVR) or use fat links through the
      `kZoneAndLinkRegistry` (~5000 flash bytes for AVR).
* Create various test objects as global variables instead of stack variables
  to get a more accurate measurement of their static memory consumption.

**v1.9.0:**
* Reduce flash usage of `BasicZoneManager` and `ExtendedZoneManager` by
  1100-1300 bytes on AVR processors:
    * Extract `BasicZoneProcessorCache` and `ExtendedZoneProcessorCache` out
      of `BasicZoneManager` and `ExtendedZoneManager`, making them
      non-templatized.
    * Remove all `virtual` methods from `ZoneManager`, making the ZoneManager
      hierarchy non-polymorphic.
    * Looks like I am reverting some of the changes made in v1.3 when I created
      the `ZoneManager` interface.
* Reduce flash usage of `BasicLinkManager` and `ExtendedLinkManager` by
  68 bytes on AVR processors by removing pure `virtual` methods on `LinkManager`
  base class.
* Increase flash usage by 34 bytes on AVR processors due to slight refactoring
  of `getHighWater()` with `getAllocSize()`. Only 4-8 bytes increase on 32-bit
  processors.

**v1.10.0:**
* Remove support for SAMD21 boards.
    * Arduino IDE 1.8.19 with SparkFun SAMD 1.8.6 can no longer upload binaries
      to these boards. Something about bossac 1.7.0 not found.
* Add memory consumption benchmarks for `ZoneSorterByName` and
  `ZoneSorterByOffsetAndName` for `BasicZoneManager` and `ExtendedZoneManager`.
    * AVR: 180-530 bytes of flash
    * 32-bit: 120-600 bytes of flash
* Upgrade tool chain:
    * Arduino IDE from 1.8.13 to 1.8.19
    * Arduino AVR from 1.8.3 to 1.8.4
    * STM32duino from 2.0.0 to 2.2.0
    * ESP32 from 1.0.6 to 2.0.2
    * Teensyduino from 1.55 to 1.56
* Add support for `fold` parameter in `LocalDateTime`, `OffsetDateTime`,
  `ZonedDateTime`, and `ExtendedZoneProcessor`. Increases flash usage:
    * AVR:
        * ~600 bytes, in `ExtendedZoneProcessor` for additional search logic,
        * ~150 bytes, `BasicZoneProcessor`, to carry along the `fold` parameter
    * most 32-bit: 400-600 bytes
    * Teensy: 1300 bytes (no idea why)

**v1.11.0**
* Upgrade ZoneInfo database so that Links are symbolic links to Zones, instead
  of hard links to Zones.
    * Allows Links to know whether they are links.
    * Allows extraction of the zoneId and zoneNames of the target Zone.
    * AVR: Increases flash consumption by ~270 bytes.
    * STM32: Increases flash by 120-150 bytes.
    * ESP8266: Increases flash by 250-300 bytes.
    * ESP32: Increases flash by ~190 bytes.
    * Teensy 3.2: Increase flash by 450-1300 bytes.

**v1.11.1**
* Change `ZoneInfoBroker::targetZoneInfo()` method to return a `ZoneInfoBroker`
  instead of raw `ZoneInfo*` pointer.
    * Increases flash usage by 4-16 bytes for the most part.

# Legend

* [1] Delta flash and ram consumption for `ZoneSorterByName` and
  `ZoneSorterByOffsetAndName` are calculated by subtracting the
  `BasicZoneManager (1 zone)` numbers, to isolate the memory consumption
  of just the sorter classes.
* [2] Delta flash and ram consumption for `ZoneSorterByName` and
  `ZoneSorterByOffsetAndName` are calculated by subtracting the
  `ExtendedZoneManager (1 zone)` numbers, to isolate the memory
  consumption of just the sorter classes.

## Arduino Nano

* 16MHz ATmega328P
* Arduino IDE 1.8.19, Arduino CLI 0.20.2
* Arduino AVR Boards 1.8.4

```
+---------------------------------------------------------------------+
| Functionality                          |  flash/  ram |       delta |
|----------------------------------------+--------------+-------------|
| baseline                               |    474/   11 |     0/    0 |
|----------------------------------------+--------------+-------------|
| LocalDateTime                          |   1132/   20 |   658/    9 |
| ZonedDateTime                          |   1306/   27 |   832/   16 |
| Manual ZoneManager                     |   1242/   13 |   768/    2 |
|----------------------------------------+--------------+-------------|
| Basic TimeZone (1 zone)                |   6534/  319 |  6060/  308 |
| Basic TimeZone (2 zones)               |   6692/  439 |  6218/  428 |
| BasicZoneManager (1 zone)              |   6740/  330 |  6266/  319 |
| BasicZoneManager (all zones)           |  19482/  706 | 19008/  695 |
| BasicZoneManager (all zones+links)     |  23814/  706 | 23340/  695 |
| BasicLinkManager (all links)           |   2378/   16 |  1904/    5 |
|----------------------------------------+--------------+-------------|
| Basic ZoneSorterByName [1]             |   6904/  330 |   370/   11 |
| Basic ZoneSorterByOffsetAndName [1]    |   7028/  330 |   494/   11 |
|----------------------------------------+--------------+-------------|
| Extended TimeZone (1 zone)             |   9900/  674 |  9426/  663 |
| Extended TimeZone (2 zones)            |  10124/ 1115 |  9650/ 1104 |
| ExtendedZoneManager (1 zone)           |  10076/  680 |  9602/  669 |
| ExtendedZoneManager (all zones)        |  31842/ 1164 | 31368/ 1153 |
| ExtendedZoneManager (all zones+links)  |  36722/ 1164 | 36248/ 1153 |
| ExtendedLinkManager (all links)        |   2570/   16 |  2096/    5 |
|----------------------------------------+--------------+-------------|
| Extended ZoneSorterByName [2]          |  10256/  680 |   356/    6 |
| Extended ZoneSorterByOffsetAndName [2] |  10378/  680 |   478/    6 |
+---------------------------------------------------------------------+

```

## Sparkfun Pro Micro

* 16 MHz ATmega32U4
* Arduino IDE 1.8.19, Arduino CLI 0.20.2
* SparkFun AVR Boards 1.1.13

```
+---------------------------------------------------------------------+
| Functionality                          |  flash/  ram |       delta |
|----------------------------------------+--------------+-------------|
| baseline                               |   3470/  153 |     0/    0 |
|----------------------------------------+--------------+-------------|
| LocalDateTime                          |   4104/  160 |   634/    7 |
| ZonedDateTime                          |   4278/  167 |   808/   14 |
| Manual ZoneManager                     |   4236/  153 |   766/    0 |
|----------------------------------------+--------------+-------------|
| Basic TimeZone (1 zone)                |   9488/  457 |  6018/  304 |
| Basic TimeZone (2 zones)               |   9648/  579 |  6178/  426 |
| BasicZoneManager (1 zone)              |   9694/  468 |  6224/  315 |
| BasicZoneManager (all zones)           |  22454/  846 | 18984/  693 |
| BasicZoneManager (all zones+links)     |  26786/  846 | 23316/  693 |
| BasicLinkManager (all links)           |   5352/  158 |  1882/    5 |
|----------------------------------------+--------------+-------------|
| Basic ZoneSorterByName [1]             |   9860/  470 |   372/   13 |
| Basic ZoneSorterByOffsetAndName [1]    |   9984/  470 |   496/   13 |
|----------------------------------------+--------------+-------------|
| Extended TimeZone (1 zone)             |  12854/  812 |  9384/  659 |
| Extended TimeZone (2 zones)            |  13080/ 1255 |  9610/ 1102 |
| ExtendedZoneManager (1 zone)           |  13030/  818 |  9560/  665 |
| ExtendedZoneManager (all zones)        |  34812/ 1302 | 31342/ 1149 |
| ExtendedZoneManager (all zones+links)  |  39692/ 1302 | 36222/ 1149 |
| ExtendedLinkManager (all links)        |   5544/  158 |  2074/    5 |
|----------------------------------------+--------------+-------------|
| Extended ZoneSorterByName [2]          |  13212/  820 |   358/    8 |
| Extended ZoneSorterByOffsetAndName [2] |  13334/  820 |   480/    8 |
+---------------------------------------------------------------------+

```

## STM32 Blue Pill

* STM32F103C8, 72 MHz ARM Cortex-M3
* Arduino IDE 1.8.19, Arduino CLI 0.20.2
* STM32duino 2.2.0

```
+---------------------------------------------------------------------+
| Functionality                          |  flash/  ram |       delta |
|----------------------------------------+--------------+-------------|
| baseline                               |  21880/ 3540 |     0/    0 |
|----------------------------------------+--------------+-------------|
| LocalDateTime                          |  22184/ 3552 |   304/   12 |
| ZonedDateTime                          |  22416/ 3568 |   536/   28 |
| Manual ZoneManager                     |  22416/ 3544 |   536/    4 |
|----------------------------------------+--------------+-------------|
| Basic TimeZone (1 zone)                |  26244/ 3708 |  4364/  168 |
| Basic TimeZone (2 zones)               |  26512/ 3872 |  4632/  332 |
| BasicZoneManager (1 zone)              |  26364/ 3728 |  4484/  188 |
| BasicZoneManager (all zones)           |  43244/ 3728 | 21364/  188 |
| BasicZoneManager (all zones+links)     |  49688/ 3728 | 27808/  188 |
| BasicLinkManager (all links)           |  23588/ 3548 |  1708/    8 |
|----------------------------------------+--------------+-------------|
| Basic ZoneSorterByName [1]             |  26536/ 3728 |   292/   20 |
| Basic ZoneSorterByOffsetAndName [1]    |  26584/ 3728 |   340/   20 |
|----------------------------------------+--------------+-------------|
| Extended TimeZone (1 zone)             |  28252/ 4096 |  6372/  556 |
| Extended TimeZone (2 zones)            |  28504/ 4648 |  6624/ 1108 |
| ExtendedZoneManager (1 zone)           |  28356/ 4104 |  6476/  564 |
| ExtendedZoneManager (all zones)        |  57588/ 4104 | 35708/  564 |
| ExtendedZoneManager (all zones+links)  |  64856/ 4104 | 42976/  564 |
| ExtendedLinkManager (all links)        |  23780/ 3548 |  1900/    8 |
|----------------------------------------+--------------+-------------|
| Extended ZoneSorterByName [2]          |  28532/ 4104 |   280/    8 |
| Extended ZoneSorterByOffsetAndName [2] |  28588/ 4104 |   336/    8 |
+---------------------------------------------------------------------+

```

An entry of `-1` indicates that the memory usage exceeded the maximum of the
microcontroller and the compiler did not generate the desired information.

## ESP8266

* NodeMCU 1.0, 80MHz ESP8266
* Arduino IDE 1.8.19, Arduino CLI 0.20.2
* ESP8266 Boards 3.0.2

```
+---------------------------------------------------------------------+
| Functionality                          |  flash/  ram |       delta |
|----------------------------------------+--------------+-------------|
| baseline                               | 260089/27892 |     0/    0 |
|----------------------------------------+--------------+-------------|
| LocalDateTime                          | 260545/27908 |   456/   16 |
| ZonedDateTime                          | 260961/27924 |   872/   32 |
| Manual ZoneManager                     | 260941/27896 |   852/    4 |
|----------------------------------------+--------------+-------------|
| Basic TimeZone (1 zone)                | 266777/28644 |  6688/  752 |
| Basic TimeZone (2 zones)               | 267145/28804 |  7056/  912 |
| BasicZoneManager (1 zone)              | 266937/28660 |  6848/  768 |
| BasicZoneManager (all zones)           | 284265/28660 | 24176/  768 |
| BasicZoneManager (all zones+links)     | 290969/28660 | 30880/  768 |
| BasicLinkManager (all links)           | 261933/27904 |  1844/   12 |
|----------------------------------------+--------------+-------------|
| Basic ZoneSorterByName [1]             | 267173/28664 |   396/   20 |
| Basic ZoneSorterByOffsetAndName [1]    | 267285/28664 |   508/   20 |
|----------------------------------------+--------------+-------------|
| Extended TimeZone (1 zone)             | 269401/29172 |  9312/ 1280 |
| Extended TimeZone (2 zones)            | 269785/29724 |  9696/ 1832 |
| ExtendedZoneManager (1 zone)           | 269529/29180 |  9440/ 1288 |
| ExtendedZoneManager (all zones)        | 299421/29176 | 39332/ 1284 |
| ExtendedZoneManager (all zones+links)  | 306989/29176 | 46900/ 1284 |
| ExtendedLinkManager (all links)        | 262125/27904 |  2036/   12 |
|----------------------------------------+--------------+-------------|
| Extended ZoneSorterByName [2]          | 269749/29184 |   348/   12 |
| Extended ZoneSorterByOffsetAndName [2] | 269845/29184 |   444/   12 |
+---------------------------------------------------------------------+

```

## ESP32

* ESP32-01 Dev Board, 240 MHz Tensilica LX6
* Arduino IDE 1.8.19, Arduino CLI 0.20.2
* ESP32 Boards 2.0.2

```
+---------------------------------------------------------------------+
| Functionality                          |  flash/  ram |       delta |
|----------------------------------------+--------------+-------------|
| baseline                               | 204501/16060 |     0/    0 |
|----------------------------------------+--------------+-------------|
| LocalDateTime                          | 206357/16188 |  1856/  128 |
| ZonedDateTime                          | 206813/16204 |  2312/  144 |
| Manual ZoneManager                     | 206829/16180 |  2328/  120 |
|----------------------------------------+--------------+-------------|
| Basic TimeZone (1 zone)                | 211361/16340 |  6860/  280 |
| Basic TimeZone (2 zones)               | 211789/16508 |  7288/  448 |
| BasicZoneManager (1 zone)              | 211525/16364 |  7024/  304 |
| BasicZoneManager (all zones)           | 228837/16364 | 24336/  304 |
| BasicZoneManager (all zones+links)     | 235541/16364 | 31040/  304 |
| BasicLinkManager (all links)           | 207709/16188 |  3208/  128 |
|----------------------------------------+--------------+-------------|
| Basic ZoneSorterByName [1]             | 211681/16364 |   320/   24 |
| Basic ZoneSorterByOffsetAndName [1]    | 211737/16364 |   376/   24 |
|----------------------------------------+--------------+-------------|
| Extended TimeZone (1 zone)             | 213773/16732 |  9272/  672 |
| Extended TimeZone (2 zones)            | 214177/17284 |  9676/ 1224 |
| ExtendedZoneManager (1 zone)           | 213893/16740 |  9392/  680 |
| ExtendedZoneManager (all zones)        | 243749/16740 | 39248/  680 |
| ExtendedZoneManager (all zones+links)  | 251317/16740 | 46816/  680 |
| ExtendedLinkManager (all links)        | 207901/16188 |  3400/  128 |
|----------------------------------------+--------------+-------------|
| Extended ZoneSorterByName [2]          | 214041/16740 |   268/    8 |
| Extended ZoneSorterByOffsetAndName [2] | 214109/16740 |   336/    8 |
+---------------------------------------------------------------------+

```

RAM usage remains constant as more objects are created, which indicates that an
initial pool of a certain minimum size is created regardless of the actual RAM
usage by objects.

## Teensy 3.2

* 96 MHz ARM Cortex-M4
* Arduino IDE 1.8.19, Arduino CLI 0.20.2
* Teensyduino 1.56

```
+---------------------------------------------------------------------+
| Functionality                          |  flash/  ram |       delta |
|----------------------------------------+--------------+-------------|
| baseline                               |  10184/ 4152 |     0/    0 |
|----------------------------------------+--------------+-------------|
| LocalDateTime                          |  10512/ 4168 |   328/   16 |
| ZonedDateTime                          |  10600/ 4184 |   416/   32 |
| Manual ZoneManager                     |  10308/ 4160 |   124/    8 |
|----------------------------------------+--------------+-------------|
| Basic TimeZone (1 zone)                |  21708/ 4324 | 11524/  172 |
| Basic TimeZone (2 zones)               |  22564/ 4488 | 12380/  336 |
| BasicZoneManager (1 zone)              |  21968/ 4344 | 11784/  192 |
| BasicZoneManager (all zones)           |  39308/ 4344 | 29124/  192 |
| BasicZoneManager (all zones+links)     |  46012/ 4344 | 35828/  192 |
| BasicLinkManager (all links)           |  11908/ 4160 |  1724/    8 |
|----------------------------------------+--------------+-------------|
| Basic ZoneSorterByName [1]             |  21832/ 4340 |   124/   16 |
| Basic ZoneSorterByOffsetAndName [1]    |  22280/ 4340 |   572/   16 |
|----------------------------------------+--------------+-------------|
| Extended TimeZone (1 zone)             |  25016/ 4712 | 14832/  560 |
| Extended TimeZone (2 zones)            |  25872/ 5264 | 15688/ 1112 |
| ExtendedZoneManager (1 zone)           |  25276/ 4720 | 15092/  568 |
| ExtendedZoneManager (all zones)        |  55100/ 4720 | 44916/  568 |
| ExtendedZoneManager (all zones+links)  |  62668/ 4720 | 52484/  568 |
| ExtendedLinkManager (all links)        |  12100/ 4160 |  1916/    8 |
|----------------------------------------+--------------+-------------|
| Extended ZoneSorterByName [2]          |  25140/ 4716 |   124/    4 |
| Extended ZoneSorterByOffsetAndName [2] |  25588/ 4716 |   572/    4 |
+---------------------------------------------------------------------+

```

